/*
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "utils/dlist.h"

#include "tdm_common.h"
#include "tdm_object.h"
#include "tdm_log.h"

#include "tdm_module.h"

#ifndef TDM_MODULES_DIR
#error  "TDM_MODULES_DIR not defined"
#endif

#ifndef TDM_MODULES_EXT
#error  "TDM_MODULES_EXT not defined"
#endif

#define TDM_MODULES_PREFIX "mod_"

/****************************************************************
 *
 ****************************************************************/

const char *tdm_module_name(const struct tdm_module *mod)
{
	return tdm_object_name(mod);
}

const char *tdm_module_description(const struct tdm_module *mod)
{
	if (!mod || !tdm_object_is_module(mod) || !mod->def)
		return NULL;

	return mod->def->description;
}

const char *tdm_module_path(const struct tdm_module *mod)
{
	if (!mod || !tdm_object_is_module(mod))
		return NULL;

	return mod->path;
}


/****************************************************************
 * Module loading + unloading
 ****************************************************************/

static char *tdm_modules_dir = TDM_MODULES_DIR;


tdm_status_t tdm_load_module(const char *name)
{
	tdm_status_t retval = TDM_STATUS_GENERR;
	struct tdm_module_def *def = NULL;
	struct tdm_module *mod = NULL;
	char path[PATH_MAX] = { '\0' };
	void *ptr = NULL;

	snprintf(path, sizeof(path), "%s/%s%s.%s", tdm_modules_dir, TDM_MODULES_PREFIX, name, TDM_MODULES_EXT);

	tdm_notice("Loading module '%s'\n", path);

	ptr = dlopen(path, RTLD_NOW | RTLD_LOCAL);
	if (ptr == NULL) {
		tdm_error("Failed to load module '%s': %s\n", path, dlerror());
		return TDM_STATUS_GENERR;
	}

	if ((def = dlsym(ptr, "tdm_module")) == NULL) {
		tdm_error("Module symbol not found: %s\n", dlerror());
		goto error;
	}

	if (!def->name) {
		tdm_error("Mandatory minimum information/functions not supplied\n");
		goto error;
	}

	if (def->name)
		tdm_info("Module name:       %s\n", def->name);
	if (def->module_load)
		tdm_info("Load function:     %p\n", def->module_load);
	if (def->module_unload)
		tdm_info("Unload function:   %p\n", def->module_unload);

	/* */
	mod = (struct tdm_module *)tdm_object_new(def->name, sizeof(*mod));
	if (!mod) {
		tdm_error("%s: failed to allocate module object\n", def->name);
		goto error;
	}

	tdm_object_set_type(mod, TDM_OBJECT_MODULE);
	mod->def  = def;
	mod->path = strdup(path);

	dlist_init(&mod->module_list);
	dlist_init(&mod->devices);

	if ((retval = tdm_register_module(mod)) != TDM_STATUS_SUCCESS) {
		tdm_error("%s: failed to register module\n", tdm_object_name(mod));
		goto error;
	}

	tdm_info("Successfully loaded module '%s'\n", name);
	return TDM_STATUS_SUCCESS;

error:
	free(mod->path);

	tdm_object_destroy((struct tdm_object *)mod);

	dlclose(ptr);
	return retval;
}

tdm_status_t tdm_load_modules(void)
{
	tdm_status_t retval = TDM_STATUS_SUCCESS;
	struct dirent entry, *result;
	int count = 0;
	char *env = NULL;
	DIR  *dir = NULL;

	/*
	 * override default modules directory (e.g. for testing)
	 * TODO: move to tdm_init_modules()
	 */
	if ((env = getenv("TDM_MODULES_DIR")) != NULL) {
		/* check if value is a directory */
		struct stat st;

		if (stat(env, &st) < 0) {
			tdm_error("Failed to stat custom modules directory '%s'\n", env);
			return TDM_STATUS_GENERR;
		}

		if (!S_ISDIR(st.st_mode)) {
			tdm_error("TDM_MODULES_DIR '%s' is not a directory\n", env);
			return TDM_STATUS_GENERR;
		}

		tdm_modules_dir = env;
	}

	/*
	 * traverse directory entries and try to load everything that
	 * looks like a "valid" module
	 */
	dir = opendir(tdm_modules_dir);
	if (!dir) {
		tdm_error("Failed to open modules directory '%s'\n", tdm_modules_dir);
		return TDM_STATUS_GENERR;
	}

	while (readdir_r(dir, &entry, &result) == 0) {
		char *extptr = NULL;
		char *name   = NULL;

		if (!result) break;

		/* only files */
		if (result->d_type != DT_REG)
			continue;

		/* no extension */
		if ((extptr = strrchr(result->d_name, '.')) == NULL)
			continue;

		/* not a module? */
		if (strncmp(result->d_name, TDM_MODULES_PREFIX, strlen(TDM_MODULES_PREFIX)) ||
		    strncmp(extptr + 1,     TDM_MODULES_EXT,    strlen(TDM_MODULES_EXT)))
			continue;

		/* prepare module name */
		name    = result->d_name + strlen(TDM_MODULES_PREFIX);
		*extptr = '\0';

		/* try to load module */
		retval = tdm_load_module(name);
		switch (retval) {
		case TDM_STATUS_SUCCESS:	/* OK */
			count++;
			break;
		default:
			tdm_error("%s: failed to load module\n", name);
		}
	}

	if (closedir(dir) < 0) {
		tdm_error("Failed to close modules directory handle\n");
	}

	if (!count) {
		tdm_notice("No modules loaded\n");
	}
	return TDM_STATUS_SUCCESS;
}

tdm_status_t tdm_set_modules_dir(const char *path)
{
	struct stat st;

	if (!path) {
		tdm_modules_dir = TDM_MODULES_DIR;
		return TDM_STATUS_SUCCESS;
	}

	if (!path[0])
		return TDM_STATUS_GENERR;

	if (stat(path, &st) < 0) {
		tdm_error("Failed to stat custom modules path '%s'\n", path);
		return TDM_STATUS_GENERR;
	}

	if (!S_ISDIR(st.st_mode)) {
		tdm_error("Custom modules path '%s' is not a directory\n", path);
		return TDM_STATUS_GENERR;
	}

	if ((tdm_modules_dir = strdup(path)) == NULL) {
		tdm_modules_dir = TDM_MODULES_DIR;
		return TDM_STATUS_GENERR;
	}

	return TDM_STATUS_SUCCESS;
}

/*************************************************************
 * Module registry
 *************************************************************/

DLIST_HEAD(tdm_module_list);
unsigned int nr_modules = 0;

/* add module to global list, after successfull load + init */
tdm_status_t tdm_register_module(struct tdm_module *mod)
{
	if (!mod || !tdm_object_is_module(mod))
		return TDM_STATUS_GENERR;

	/* TODO: lock tdm_module_list */

	dlist_insert_tail(&tdm_module_list, &mod->module_list);
	nr_modules++;

	/* TODO: unlock tdm_module_list */
	return TDM_STATUS_SUCCESS;
}

/* remove module from global list */
tdm_status_t tdm_unregister_module(struct tdm_module *mod)
{
	if (!mod || !tdm_object_is_module(mod))
		return TDM_STATUS_GENERR;

	/* TODO: lock tdm_module_list */

	dlist_remove(&mod->module_list);
	nr_modules--;

	/* TODO: unlock tdm_module_list */
	return TDM_STATUS_SUCCESS;
}

struct tdm_module *tdm_find_module(const char *name)
{
	struct tdm_module *mod = NULL;
	struct dlist_head *entry;

	if (!nr_modules)
		return NULL;

	dlist_for_each(&tdm_module_list, entry) {
		mod = dlist_entry(entry, struct tdm_module, module_list);

		if (!tdm_object_is_module(mod)) {
			tdm_error("%p: object is not a module\n", mod);
			return NULL;
		}

		if (!strcmp(name, tdm_object_name(mod)))
			break;
	}
	return mod;
}

/* DEBUG */
void tdm_list_modules(void)
{
	struct dlist_head *entry;

	tdm_printf("-------------------------------- modules loaded ---------------------------------\n");

	if (!nr_modules)
	tdm_printf("\n                        - no modules have been loaded -\n\n");

	dlist_for_each(&tdm_module_list, entry) {
		struct tdm_module *mod = dlist_entry(entry, struct tdm_module, module_list);

		if (!tdm_object_is_module(mod)) {
			tdm_error("%p: object is not a module\n", mod);
			continue;
		}

		const char *desc = tdm_module_description(mod);
		const char *path = tdm_module_path(mod);

		tdm_printf("+ %s:\n", tdm_object_name(mod));
		tdm_printf("\tDescription:    %-50s\n", (desc) ? desc : "-");
		tdm_printf("\tPath:           %-50s\n", (path) ? path : "-");
		tdm_printf("\tUse count:      %d\n", tdm_object_use_count(mod));
		tdm_printf("\tDevices:        %d\n", mod->nr_devices);
	}

	tdm_printf("---------------------------------------------------------------------------------\n");
}
