/*
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>

#include "utils/dlist.h"

#include "tdm_common.h"
#include "tdm_object.h"
#include "tdm_log.h"

#include "tdm_port.h"
#include "tdm_channel.h"

struct tdm_port *tdm_object_to_port(struct tdm_object *obj)
{
	if (!obj || !tdm_object_is_port(obj))
		return NULL;

	return (struct tdm_port *)obj;
}

struct tdm_module *tdm_port_get_module(struct tdm_port *port)
{
	if (!port || !tdm_object_is_port(port))
		return NULL;

	return (struct tdm_module *)tdm_object_owner(port);
}

struct tdm_port *tdm_port_new(const char *name, const struct tdm_port_ops *ops)
{
	struct tdm_port *port = NULL;

	if (!name || !name[0] || !ops)
		return NULL;

	port = (struct tdm_port *)tdm_object_new(name, sizeof(*port));
	if (!port)
		return NULL;

	tdm_object_set_type(port, TDM_OBJECT_PORT);

	port->ops = (struct tdm_port_ops *)ops;

	return port;
}

tdm_status_t tdm_port_destroy(struct tdm_port *port)
{
	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	/* TODO: check if port is still in use */
	if (port->nr_channels) {
		tdm_error("%s: port still has channels\n", tdm_object_name(port));
		return TDM_STATUS_GENERR;
	}

	if (port->ops->port_close) {
		tdm_status_t retval;
		if ((retval = port->ops->port_close(port)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to destroy port\n", tdm_object_name(port));
			return retval;
		}
	} else {
		tdm_debug("%s: no port destroy callback\n", tdm_object_name(port));
	}

	tdm_object_destroy((struct tdm_object *)port);
	return TDM_STATUS_SUCCESS;
}

tdm_status_t tdm_port_add_channel(struct tdm_port *port, struct tdm_channel *chan)
{
	struct tdm_module *owner = NULL;
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	owner = tdm_port_get_module(port);
	if (!owner) {
		tdm_error("%s: orphaned port", tdm_object_name(port));
		return TDM_STATUS_GENERR;
	}

	/* TODO: lock port */

	dlist_insert_tail(&port->channels, &chan->port_channel_list);
	port->nr_channels++;

	chan->port = port;

	tdm_object_set_owner(chan, owner);

	/* TODO: unlock port */

	/* */
	if (port->ops->port_add_channel) {
		if ((retval = port->ops->port_add_channel(port, chan)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to add channel \"%s\"\n", tdm_object_name(port), tdm_object_name(chan));
			goto error;
		}
	} else {
		tdm_debug("%s: no add channel callback\n", tdm_object_name(port));
	}

	return TDM_STATUS_SUCCESS;

error:
	dlist_remove(&chan->port_channel_list);
	port->nr_channels--;
	chan->port = NULL;

	return retval;
}

int tdm_port_has_channel(struct tdm_port *port, struct tdm_channel *chan)
{
	if (!port || !tdm_object_is_port(port))
		return TDM_FALSE;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_FALSE;

	return (dlist_get_head(&chan->port_channel_list) == dlist_get_head(&port->channels));
}

tdm_status_t tdm_port_remove_channel(struct tdm_port *port, struct tdm_channel *chan)
{
	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	if (!tdm_port_has_channel(port, chan)) {
		tdm_error("%s: not a member of port \"%s\"", tdm_object_name(chan), tdm_object_name(port));
		return TDM_STATUS_GENERR;
	}

	/* */
	if (port->ops->port_remove_channel) {
		tdm_status_t retval;
		if ((retval = port->ops->port_remove_channel(port, chan)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to remove channel \"%s\"\n", tdm_object_name(port), tdm_object_name(chan));
			return retval;
		}
	} else {
		tdm_debug("%s: no remove channel callback\n", tdm_object_name(port));
	}


	/* TODO: lock port */

	dlist_remove(&chan->port_channel_list);
	port->nr_channels--;
	chan->port = NULL;

	/* TODO: unlock port */

	return TDM_STATUS_SUCCESS;
}
