/*
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>

#include "utils/dlist.h"

#include "tdm_common.h"
#include "tdm_object.h"
#include "tdm_log.h"

#include "tdm_span.h"
#include "tdm_spangroup.h"

#include "tdm_channel.h"
#include "tdm_event.h"

struct tdm_spangroup *tdm_span_get_spangroup(struct tdm_span *span)
{
	struct tdm_spangroup *group = NULL;
	struct dlist_head *entry;

	if (!span || !tdm_object_is_span(span))
		return NULL;

	entry = dlist_get_head(&span->group_span_list);
	if (!entry) {
		tdm_error("%s: span not member of a group\n", tdm_object_name(span));
		return NULL;
	}

	group = dlist_entry(entry, struct tdm_spangroup, spans);
	if (!tdm_object_is_spangroup(group)) {
		tdm_error("%p: object is not a span group\n", group);
		return NULL;
	}

	return group;
}

struct tdm_span *tdm_object_to_span(struct tdm_object *obj)
{
	if (!obj || !tdm_object_is_span(obj))
		return NULL;

	return (struct tdm_span *)obj;
}

struct tdm_span *tdm_span_new(const char *name)
{
	struct tdm_span *span = NULL;

	span = (struct tdm_span *)tdm_object_new(name, sizeof(*span));
	if (!span)
		return NULL;

	tdm_object_set_type(span, TDM_OBJECT_SPAN);

	dlist_init(&span->channels);

	return span;
}

tdm_status_t tdm_span_add_channel(struct tdm_span *span, struct tdm_channel *chan)
{
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	if (tdm_channel_has_span(chan)) {
		struct tdm_span *owner = tdm_channel_get_span(chan);
		tdm_error("%s: already assigned to span \"%s\"\n", tdm_object_name(chan), tdm_object_name(owner));
		return TDM_STATUS_GENERR;
	}

	/* TODO: locking */

	dlist_insert_tail(&span->channels, &chan->span_channel_list);
	span->nr_channels++;
	chan->span = span;

	/* TODO: locking */

	/* let signalling module handle span add operation
	 * TODO: do this before, of after adding the channel to the span??
	 */
	if (span->ops->span_add_channel) {
		if ((retval = span->ops->span_add_channel(span, chan)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to remove channel \"%s\"\n", tdm_object_name(span), tdm_object_name(chan));
			goto error;
		}
	} else {
		tdm_debug("%s: no add channel callback\n", tdm_object_name(span));
	}

	return TDM_STATUS_SUCCESS;

error:
	dlist_remove(&chan->span_channel_list);
	span->nr_channels--;
	chan->span = NULL;
	return retval;
}

int tdm_span_has_channel(struct tdm_span *span, struct tdm_channel *chan)
{
	if (!span || !tdm_object_is_span(span))
		return TDM_FALSE;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_FALSE;

	return (dlist_get_head(&chan->span_channel_list) == dlist_get_head(&span->channels));
}

tdm_status_t tdm_span_remove_channel(struct tdm_span *span, struct tdm_channel *chan)
{
	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	if (!tdm_span_has_channel(span, chan)) {
		tdm_error("%s: not a member of span \"%s\"\n", tdm_object_name(chan), tdm_object_name(span));
		return TDM_STATUS_GENERR;
	}

	/* TODO: more checks? e.g. channel in use? */

	/* see if signalling module allows removing the channel */
	if (span->ops->span_remove_channel) {
		tdm_status_t retval;
		if ((retval = span->ops->span_remove_channel(span, chan)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to remove channel \"%s\"\n", tdm_object_name(span), tdm_object_name(chan));
			return retval;
		}
	} else {
		tdm_debug("%s: no remove channel callback\n", tdm_object_name(span));
	}

	/* TODO: locking */

	dlist_remove(&chan->span_channel_list);
	span->nr_channels--;
	chan->span = NULL;

	/* TODO: locking */

	return TDM_STATUS_SUCCESS;
}



/******************************************************************************
 * Event handling
 ******************************************************************************/

/**
 * callback to destroy span event
 */
static void tdm_span_destroy_event(struct tdm_event *event)
{
	if (tdm_event_class(event) != TDM_EVENT_SPAN) {
		tdm_error("%p: is not a span event\n", event);
		return;
	}

	if (!tdm_object_is_span(event->ev_obj.span)) {
		tdm_error("%p: object is not a span\n", event->ev_obj.span);
		return;
	}

	tdm_object_put(event->ev_obj.span);

	/* clean up per-event data */
}

tdm_status_t tdm_span_queue_event(struct tdm_span *span, const tdm_span_event_t ev_type, const void *data)
{
	struct tdm_event *event = NULL;
	tdm_status_t retval;

	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	event = tdm_event_new(TDM_EVENT_SPAN);
	if (!event) {
		tdm_error("%s: failed to allocate new event %d\n", tdm_object_name(span), ev_type);
		return TDM_STATUS_GENERR;
	}

	event->ev_obj.span = tdm_object_get(span);
	event->ev_destroy  = &tdm_span_destroy_event;

	/* add per-event data? */
	switch (ev_type) {
	case TDM_SPAN_EVENT_SPAN_READY:
	case TDM_SPAN_EVENT_SPAN_STOPPED:
		break;
	default:
		if (data)
			event->ev_data.custom = (void *)data;
	}

	retval = tdm_event_queue(event);
	if (retval != TDM_STATUS_SUCCESS) {
		tdm_error("%s: failed to enqueue new event %d\n", tdm_object_name(span), ev_type);
		goto error;
	}

	return TDM_STATUS_SUCCESS;

error:
	tdm_span_destroy_event(event);
	return retval;
}


/**************************************************************************
 * Other
 **************************************************************************/

tdm_span_state_t tdm_span_get_state(struct tdm_span *span)
{
	if (!span || !tdm_object_is_span(span))
		return TDM_SPAN_STATE_INVALID;

	return span->state;
}

tdm_status_t tdm_span_start(struct tdm_span *span)
{
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!span || !tdm_object_is_span(span) || !span->ops)
		return TDM_STATUS_GENERR;

	/*
	 * Call (signalling) module's span start handler
	 */
	if (span->ops->span_start) {
		retval = span->ops->span_start(span);
		if (retval != TDM_STATUS_SUCCESS) {
//			tdm_error("%s: error starting span: %s\n", tdm_object_name(span), tdm_errstr(ret));
			tdm_error("%s: error starting span\n", tdm_object_name(span));
			goto out;
		}
	}

	/* common span start */

#ifdef __TODO__
	pthread_create();

	/* TODO: span thread should set state to up/active and send SPAN_READY event */
#endif
out:
	return retval;
}

/*
 * possible change:
 *
 * typedef enum {
 * 	TDM_SHUTDOWN_GRACEFUL = 0,  // wait for all calls to complete (may block for a long time...)
 *	TDM_SHUTDOWN_IMMEDIATE      // kill all calls, shutdown span
 * } tdm_shutdown_type;
 *
 * #define TDM_SHUTDOWN_DEFAULT	TDM_SHUTDOWN_IMMEDIATE	// default: immediate
 *
 * tdm_span_stop(struct tdm_span *span, tdm_shutdown_type how)
 */

tdm_status_t tdm_span_stop(struct tdm_span *span)
{
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!span || !tdm_object_is_span(span) || !span->ops)
		return TDM_STATUS_GENERR;
#if __TODO__
	/*
	 * 1. put span into non-active state (= not accepting new calls)
	 *
	 * 2. for immediate:
	 *	- kill all active calls
	 *	- invoke callbacks
	 *	- mark span down/stopped
	 *	- send SPAN_STOPPED span event to user
	 *
	 *   for graceful:
	 *	- schedule callback for "all calls have ended"
	 *	- in callback:
	 *	- invoke span callbacks
	 *	- mark span down/stopped
	 *	- send SPAN_STOPPED span event to user
	 *
	 */
#endif

	/*
	 * Call (signalling) module's span stop handler
	 */
	if (span->ops->span_stop) {
		retval = span->ops->span_stop(span);
		if (retval != TDM_STATUS_SUCCESS) {
//			tdm_error("%s: error stopping span: %s\n", tdm_object_name(span), tdm_errstr(ret));
			tdm_error("%s: error stopping span\n", tdm_object_name(span));
			goto out;
		}
	}

	/* common span stop */
#ifdef __TODO__
	pthread_kill();
#endif

out:
	return retval;
}
