#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>

#include <dahdi/user.h>

/*
 * public api
 */
typedef enum {
	TDM_STATUS_SUCCESS = 0,
	TDM_STATUS_GENERR,
	TDM_STATUS_INVALID,
	TDM_STATUS_UNIMPLEMENTED,
	TDM_STATUS_NOTFOUND,
} tdm_status_t;

typedef enum {
	TDM_LOG_NONE = 0,
	TDM_LOG_EMERGENCY,
	TDM_LOG_CRITICAL,
	TDM_LOG_ALERT,
	TDM_LOG_ERROR,
	TDM_LOG_WARNING,
	TDM_LOG_NOTICE,
	TDM_LOG_INFO,
	TDM_LOG_DEBUG,
	TDM_LOG_INVALID
} tdm_loglevel_t;

/* public version */
/*#define TDM_API*/

/*
 * internal, tdm_priv.h
 */
#if defined(HAVE_GCC_VISIBILITY) || defined(HAVE_SUNCC_VISIBILITY)
#define TDM_API	__attribute__((visibility("default")))
#else
#define TDM_API
#endif

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(x)	(sizeof(x) / sizeof(x[0]))
#endif

#ifndef ACCESS_ONCE
#define ACCESS_ONCE(x)	(*(volatile typeof(x) *)&(x))
#endif

#ifndef offset_of
#define offset_of(type, member)	\
	(unsigned long)&(((type *)0)->member)
#endif

#ifndef container_of
#define container_of(x, type, member)	\
	((type *)((unsigned long)(x) - offset_of(type, member)))
#endif

#if defined(__GNUC__)
#define likely(x)	__builtin_expect(!!(x), 1)
#define unlikely(x)	__builtin_expect(!!(x), 0)
#else
#define likely(x)	(x)
#define unlikely(x) 	(x)
#endif


/*
 * internal, tdm_log.c
 */
const static struct tdm_loglevel_name {
	const tdm_loglevel_t id;
	const char *name;
} tdm_loglevel_names[] = {
	{ TDM_LOG_NONE,      "none"      },
	{ TDM_LOG_EMERGENCY, "emergency" },
	{ TDM_LOG_CRITICAL,  "critical"  },
	{ TDM_LOG_ALERT,     "alert"     },
	{ TDM_LOG_ERROR,     "error"     },
	{ TDM_LOG_WARNING,   "warning"   },
	{ TDM_LOG_NOTICE,    "notice"    },
	{ TDM_LOG_INFO,      "info"      },
	{ TDM_LOG_DEBUG,     "debug"     },
};
const static int tdm_loglevel_names_count = ARRAY_SIZE(tdm_loglevel_names);

/* public api */
TDM_API const char *
tdm_loglevel_get_name(const tdm_loglevel_t level)
{
	for (int i = 0; i < tdm_loglevel_names_count; i++) {
		if (tdm_loglevel_names[i].id == level) {
			return tdm_loglevel_names[i].name;
		}
	}
	return NULL;
}

TDM_API const tdm_loglevel_t *
tdm_loglevel_get_id(const char *name)
{
	for (int i = 0; i < tdm_loglevel_names_count; i++) {
		if (strcmp(tdm_loglevel_names[i].name, name) == 0) {
			return tdm_loglevel_names[i].id;
		}
	}
	return TDM_LOG_INVALID;
}

/* private tdm_log.c */
static struct tdm_log_global {
	tdm_loglevel_t level;
} tdm_log_global;

/* public */
TDM_API tdm_status_t
tdm_set_loglevel(const tdm_loglevel_t level)
{
	if (level < TDM_LOG_NONE || level >= TDM_LOG_INVALID)
		return TDM_STATUS_INVALID;

	tdm_log_global.level = level;
	return TDM_STATUS_SUCCESS;
}

TDM_API tdm_loglevel_t
tdm_get_loglevel(void)
{
	return tdm_log_global.level;
}

/* public (maybe only modules api) */
TDM_API tdm_status_t
__tdm_log(const loglevel_t level, const void *p1, const void *p2, const char *fmt, ...)
{
	va_list ap;

	if (level > tdm_log_global.level)	/* ignore everything above current level */
		return TDM_STATUS_SUCCESS;

	va_start(ap, fmt);
	vsprintf(fmt, ap);
	va_end(ap);
	return TDM_STATUS_SUCCESS;
}

#define tdm_log(level, fmt, args...) \
	__tdm_log(level, NULL, NULL, fmt, ##args)

/* public (maybe only modules api), convenience macros */
#define tdm_info(fmt, args...)	\
	__tdm_log(TDM_LOG_INFO, NULL, NULL, fmt, ##args)

#define tdm_error(fmt, args...)	\
	__tdm_log(TDM_LOG_ERROR, NULL, NULL, fmt, ##args)

#define tdm_warn(fmt, args...)	\
	__tdm_log(TDM_LOG_WARNING, NULL, NULL, fmt, ##args)

#define tdm_debug(fmt, args...)	\
	__tdm_log(TDM_LOG_DEBUG, NULL, NULL, fmt, ##args)


/**
 * Module definition (api, only visible part of a module, const)
 */
struct tdm_module_handle;

struct tdm_module {
	/* Common */
	const char *name;
	tdm_status (*init)(const struct tdm_module_handle *mod);	/*!< Module init */
	tdm_status (*fini)(const struct tdm_module_handle *mod);	/*!< Module exit */
	tdm_status (*reload)(const struct tdm_module_handle *mod);	/*!< Re-init module */

	/* I/O Module-specific  */
	tdm_status (*probe)(const struct tdm_module_handle *mod);	/*!< I/O modules only: detect hardware */

	/* SIG Module-specific */
};

#if 0
/* internal part */
struct tdm_module_handle {
	struct dlist_head list;
	const struct tdm_module *mod;
	/* ... */
};
#endif

const static char* dahdi_ctl_path = "/dev/dahdi/ctl";

/**
 * globals
 */
static struct dahdi_globals {
	int ctl_fd;	/*!< control device fd */
} dahdi_globals;

/**
 * Module init function
 */
static tdm_status_t dahdi_init(const struct tdm_module_handle *module)
{
	struct dahdi_versioninfo dversion;
	tdm_status_t retval = TDM_STATUS_GENERR;
	int fd  = 0;
	int res = 0;

	fd = open(dahdi_ctl_path, O_RDWR);
	if (fd < 0) {
		tdm_error("Failed to open DAHDI control device \"%s\": %s\n", dahdi_ctl_path, strerror());
		return TDM_STATUS_GENERR;
	}

	res = ioctl(fd, DAHDI_GETVERSION, &dversion);
	if (res < 0) {
		tdm_error("Failed to retrieve DAHDI version: %s\n", strerror());
		goto out;
	}

	tdm_info("DAHDI Version: %s\n", dversion.version);

	memset(&dahdi_globals, 0, sizeof(dahdi_globals));
	dahdi_globals.ctl_fd = fd;


	tdm_debug("Successfully initialized DAHDI module, hardware will be probed later\n");
#if 0
	/* register interfaces */
	tdm_register_sigmod();
	tdm_register_iomod();
#endif

out:
	return retval;
}

/**
 * Module exit function
 */
static tdm_status_t dahdi_fini(const struct tdm_module_handle *module)
{
	return TDM_STATUS_SUCCESS;
}

/**
 * Module reload function
 */
static tdm_status_t dahdi_reload(const struct tdm_module_handle *module)
{
	return TDM_STATUS_UNIMPLEMENTED;
}

/**
 * Hardware probing / device detection
 * (This is an I/O module after all)
 */
static tdm_status_t dahdi_probe(const struct tdm_module_handle *module)
{
	tdm_status_t retval = TDM_STATUS_GENERR;
	struct dahdi_spaninfo dspan;
	int nr_channels = 0;
	int nr_spans = 0;

	for (int x = nr_spans = 1; x <= nr_spans; x++) {
		memset(&dspan, 0, sizeof(struct dahdi_spaninfo));
		dspan.spanno = x;

		res = ioctl(fd, DAHDI_SPANSTAT, &dspan);
		if (res < 0) {
			tdm_warning("dahdi: Failed to get status of span %d\n", x);
			break;
		}

		/*
		 * TODO: tdm_logm_begin(TDM_LOG_INFO) -> tdm_logm_append("....", xyz)...tdm_logm_end() / tdm_logm_commit() / tdm_logm_discard()
		 *
		 * - multi-line message logging api (use parts of code in libisdn)
		 *   (use buffer-chaining?  max message size limit?)
		 *
		 * - speed-up: do not allocate buffer if current loglevel is too low
		 *   (and skip writing stuff into it, turning tdm_logm_append() + tdm_logm_commit() into a no-op)
		 */
		printf("Span %d status:\n", dspan.spanno);
		printf("   Name:        %s\n", dspan.name);
		printf("   Description: %s\n", dspan.desc);
		printf("   Channels #:  %d\n", dspan.numchans);
		printf("   Channels #:  %d\n", dspan.totalchans);
		puts("");

		/* for dahdi:
		 *
		 * - need to parse span name (<driver>/<card>/<port>? e.g. WCTDM/1/0)
		 *   to get device (card) <-> port association (let's hope this is sane on DAHDI's side)
		 *
		 * - a way to track already registered devices, so we can do:
		 *
		 *	if (card_registered(...)) {
		 *		tdm_device_add_port(...);
		 *	} else {
		 *		tdm_device_register(...);
		 *		tdm_device_add_port(...);
		 *	}
		 *
		 *   on second thought, maybe add some kind of tdm_device_get_by_name()
		 *   and let the i/o module register them with a (custom) name like "dahdi:%span_id%"
		 *   (maybe add the prefix internally), this will turn the code into:
		 *
		 *	if ((device = tdm_device_get_by_name("dahdi:%span_id%")) == NULL) {
		 *		tdm_device_register(module, device, ...);
		 *	}
		 *	tdm_device_add_port(...);
		 *
		 *   higher level functions could easily use "module:%card_id%:%port_id%"
		 */

		nr_spans     = dspan.totalspans;
		nr_channels += dspan.totalchans;
	}
	if (nr_spans == 0) {
		tdm_warning("dahdi: No spans found");
		retval = TDM_STATUS_NOTFOUND;
		goto out;
	}

	tdm_info("dahdi: %d span(s), %d channel(s) found\n", nr_spans, nr_channels);

out:
	return retval;
}


/*
 *
 */
const struct tdm_module module = {
	.name   = "dahdi",
	.init   = dahdi_init,
	.fini   = dahdi_fini,
	.reload = dahdi_reload,
	.probe  = dahdi_probe
};




/*
 * open dahdi ctl interface
 */
#if 0
int main(void)
{
	struct dahdi_params dparams;
	struct dahdi_spaninfo dspan;
	int nr_spans = 0, nr_channels = 0;
	int fd  = 0;
	int res = 0;


/*
	memset(&dparams, 0, sizeof(struct dahdi_params));
	dparams.spanno = 1;

	res = ioctl(fd, DAHDI_GET_PARAMS, &dparams);
	if (res < 0) {
		perror("asking for span 1 configuration");
		goto out;
	}

	puts("Span 1 configuration:\n");
	printf("   SigType: %d\n", dparams.sigtype);
*/
	for (int x = nr_spans = 1; x <= nr_spans; x++) {
		memset(&dspan, 0, sizeof(struct dahdi_spaninfo));
		dspan.spanno = x;

		res = ioctl(fd, DAHDI_SPANSTAT, &dspan);
		if (res < 0) {
			perror("asking for span information");
			break;
		}

		printf("Span %d status:\n", dspan.spanno);
		printf("   Name:        %s\n", dspan.name);
		printf("   Description: %s\n", dspan.desc);
		printf("   Channels #:  %d\n", dspan.numchans);
		printf("   Channels #:  %d\n", dspan.totalchans);
		puts("");

		nr_spans     = dspan.totalspans;
		nr_channels += dspan.totalchans;
	}
	if (nr_spans == 0) {
		puts("dahdi: No spans found");
		res = ENODEV;
	} else {
		printf("dahdi: %d span(s), %d channel(s) found\n", nr_spans, nr_channels);
	}

out:
	close(fd);
	return res;
}
#endif
